
/* 
 * Recursive Binary Inertial Partition
 *
 * Purpose
 *   Divides the object perpendicular to the long axis of the object
 *   as determined by the relative magnitude of the moment of inertia
 *   along each principal axis.
 *
 * Input
 *   ndata is the number of elements (nodes)
 *   x[] is the list of X coordinates
 *   y[] is the list of Y coordinates
 *   z[] is the list of Z coordinates
 *   level is the level of recursion, i.e. the domain will be
 *	divided into (2 ^ level) subdomains
 *
 * Output
 *   no output is returned
 *
 * Note
 *   coordinates x[], y[], and z[] are redistributed based on
 *	the result of binary inertial partitioning
 */

#include <stdio.h>
#include <math.h>
#include "parti.h"
#include "bisec.h"

#define  MAXPART	512	/* maximum number of partitions */
#define	 REMAPLEVEL	2	/* default value */
#define	 NDIM		4	/* maximum dimenstions */
#define	 NTENSOR	10	/* maximum size of each inertial tensor
				   (4x4 symmetric matrix) */

extern	double *doublelist;	/* array of double numbers */

void
RecInerBisecMap (level, maparray, ndata, ndim, data) 
int    level;			/* level of binary disection */
int    *maparray;		/* result mapping of the partition */
int    ndata;			/* number of nodes */
int    ndim;			/* number of dimensions */
double *data[];    		/* coordinate arrays */

{
  int    totalpart;		/* total number of partitions */
  int    npart;			/* current number of partitions */
  int    parproc;		/* # of processors of each partition */
  int    mask;			/* mask */
  int    ptr[MAXPART+1];	/* pointer of begin of partition */
  int    gsize[MAXPART];	/* size of each partition */
  int    work[MAXPART];		/* working array */
  int    *proc;			/* processor assignments */
  int    *index;		/* index of arrays */
  int    i, j, k, l, m, n, no;
  double C[MAXPART][NDIM];	/* array of centers of masses */
  double V[MAXPART][NDIM];	/* array of eigenvectors */
  double M[MAXPART][NTENSOR];	/* array of inertial tensors */
  double dwork[MAXPART][NTENSOR];/* double working array */
  double *dot;			/* array of inner product */
  Val    *midlist;		/* array of medians */
  Val    *temp;
  SCHED  *sched;		/* schedule for remapping */
  TTABLE *table;		/* translation table */
  TTABLE *init_ttable_with_proc();
  int	 comparedouble();

  /* initialize parameters */
  totalpart = (1 << level);

  /* allocate memory for necessary arrays*/
  index   = (int *) malloc ((ndata+PARTI_numnodes) * sizeof(int));
  proc    = (int *) malloc ((ndata+PARTI_numnodes) * sizeof(int));
  dot     = (double *) malloc ((ndata+PARTI_numnodes) * sizeof(double));
  midlist = (Val *) malloc (totalpart * PARTI_numnodes * sizeof(Val));
  temp    = (Val *) malloc (totalpart * PARTI_numnodes * sizeof(Val));
  doublelist  = dot;

  /* put the coordinates into data array */
  for (i = 0; i <= ndata; i++) {
    index[i] = i;
    proc[i] = 0;
  }

  /* sort the arrays of each partition and find the median */
  for (i=0, ptr[0]=0; i < level; i++) {
    /* find the global sizes of partitions */
    npart = (1 << i);			/* number of partitions */
    ptr[npart] = ndata;    		/* number of nodes */
    for (j = 0; j < npart; j++)
      gsize[j] = ptr[j+1] - ptr[j];
    addivec (level, gsize, npart, work);

    /* find the center of mass of each partition */
    for (j = 0; j < npart; j++)
      for (k = 0; k < ndim; k++)
        C[j][k] = 0.0;
    for (j = 0; j < npart; j++) {
      for (k = ptr[j]; k < ptr[j+1]; k++) {
	n = index[k];
	for (m = 0; m < ndim; m++)
          C[j][m] += data[m][n];
      }
    }
    adddvec (level, C, npart*NDIM, dwork);
    for (j = 0; j < npart; j++)
      for (k = 0; k < ndim; k++)
    	C[j][k] /= gsize[j];

#ifdef	DEBUG
    if (PARTI_mynode == 0)
      printf ("Center of mass: %f %f %f\n", C[0][0], C[0][1], C[0][2]);
#endif

    /* calculate the moment of inertial tensor */
    for (j = 0; j < npart; j++)
      for (k = 0; k < NTENSOR; k++)
    	M[j][k] = 0.0;
    for (j = 0; j < npart; j++) {
      for (k = ptr[j]; k < ptr[j+1]; k++) {
	n = index[k];
	for (no = 0, l = 0; l < ndim; l++)
	  for (m = l; m < ndim; no++, m++)
            M[j][no] += (data[l][n] - C[j][l]) * (data[m][n] - C[j][m]);
      }
    }
    adddvec (level, M, npart*NTENSOR, dwork);
#ifdef	DEBUG
    if (PARTI_mynode == 0) {
      printf ("Moment of Inertial Tensor:\n");
      printf ("\t%f\t%f\t%f\n", M[0][0], M[0][1], M[0][2]);
      printf ("\t%f\t%f\t%f\n", M[0][1], M[0][3], M[0][4]);
      printf ("\t%f\t%f\t%f\n", M[0][2], M[0][4], M[0][5]);
    }
#endif

    /* find the eigenvalues of eigenvectors */
    for (j = 0; j < npart; j++)
      find_eigenvector (ndim, &M[j][0], &V[j][0]);

#ifdef	DEBUG
    if (PARTI_mynode == 0)
      printf ("Eigenvector: %f %f %f\n", V[0][0], V[0][1], V[0][2]);
#endif

    /* projection (inner product) of each element to eigenvector */
    for (j = 0; j < npart; j++) {
      for (k = ptr[j]; k < ptr[j+1]; k++) {
        n = index[k];
	for (dot[n] = 0, m = 0; m < ndim; m++)
          dot[n] += V[j][m]*(data[m][n]-C[j][m]);
      }
    }

    /* sort the dot product array of each partition */
    for (j = 0; j < npart; j++)
      qsort (&index[ptr[j]], ptr[j+1]-ptr[j], sizeof(int), comparedouble);

    /* apply the method median of medians to find 
       the median of the distributed array */
    find_median (level, npart, ptr, dot, index, proc, midlist, temp);
  }

  /* put the result of partition to maparray */
  for (i = 0; i < ndata; i++)
    maparray[index[i]] = proc[index[i]];

  /* release memory */
  free (index);
  free (proc);
  free (dot);
  free (temp);
  free (midlist);
}

/* when the number of processor is large, the calls to InerBisecMap()
   will be changed to InerBisec() to reduce execution time. */
void
RecInerBisecRemap (level, maparray, ndata, ndim, data)
int	level;			/* level of binary disection */
int     *maparray;              /* result of partitioning */
int     ndata;                  /* number of nodes */
int	ndim;			/* number of dimenstions */
double  *data[];		/* coordinates */

{
    int     start;          /* start index of local data */
    int     newndata;       /* number of data after partitioning */
    int     nlist[MAXPART]; /* list of ndata */
    int     temp[MAXPART];  /* temporary array */
    int     *myindex;       /* indexes of local elements */
    int     totalsize, localsize;
    int     *proc, *local;
    int     i, j;
    double  **newdata;
    TTABLE  *table;
    SCHED   *sched;

    /* find the global size */
    for (i = 0; i < PARTI_numnodes; i++)
        nlist[i] = 0;
    nlist[PARTI_mynode] = ndata;
    PARTI_gisum (nlist, PARTI_numnodes, temp);
    for (localsize = nlist[0], i = 1; i < PARTI_numnodes; i++) {
	if (nlist[i] > localsize)
	    localsize = nlist[i];
        nlist[i] += nlist[i-1];
    }
    totalsize = nlist[PARTI_numnodes-1];
    localsize = ((localsize << 1) > totalsize)? totalsize: (localsize << 1);

    /* allocate memory */
    myindex = (int *) malloc (localsize * sizeof(int));
    newdata = (double **) malloc (sizeof(double *) * ndim);
    for (i = 0; i < ndim; i++)
        newdata[i] = (double *) malloc (localsize * sizeof(double));

    /* generate indexes */
    start = (PARTI_mynode) ? nlist[PARTI_mynode-1]: 0;
    for (i = 0; i < ndata; i++)
        myindex[i] = start + i;

    /* build translation table */
    table = build_translation_table (BLOCKED, myindex, ndata);

    /* copy data arrays to new arrays */
    for (i = 0; i < ndim; i++)
        for (j = 0; j < ndata; j++) {
    	    newdata[i][j] = data[i][j];
    }

    /* call the binary disection */
    newndata = ndata;
    RecInerBisec (level, REMAPLEVEL, myindex, &newndata, 3, newdata);

    /* free allocated memory */
    for (i = 0; i < ndim; i++)
        free (newdata[i]);
    free (newdata);

    /* generate a schedule */
    proc  = (int *) malloc (newndata * sizeof(int));
    local = (int *) malloc (newndata * sizeof(int));
    dereference (table, myindex, local, proc, newndata);
    sched = schedule (local, proc, newndata, 1);

    /* produce maparray */
    for (i = 0; i < newndata; i++)
        proc[i] = PARTI_mynode;
    PARTI_scatter (sched, proc, maparray, sizeof(int), NULL);

    /* free allocated memory */
    free (local);
    free (proc);
    free (myindex);
    free_table (table);
    free_sched (sched);
}

void
WeighRecInerBisecMap (level, maparray, load, ndata, ndim, data) 
int    level;			/* level of binary disection */
int    *maparray;		/* result mapping of the partition */
int    *load;			/* load of each element */
int    ndata;			/* number of nodes */
double *data[];    		/* coordinate arrays */

{
  int    totalpart;		/* total number of partitions */
  int    npart;			/* current number of partitions */
  int    parproc;		/* # of processors of each partition */
  int    mask;			/* mask */
  int    ptr[MAXPART+1];	/* pointer of begin of partition */
  int    gsize[MAXPART];	/* size of each partition */
  int    work[MAXPART];		/* working array */
  int    *proc;			/* processor assignments */
  int    *index;		/* index of arrays */
  int    *weight;		/* weight array */
  int    i, j, k, l, m, n, no;
  double C[MAXPART][NDIM];	/* array of centers of masses */
  double V[MAXPART][NDIM];	/* array of eigenvectors */
  double M[MAXPART][NTENSOR];	/* array of inertial tensors */
  double dwork[MAXPART][NTENSOR];/* double working array */
  double *dot;			/* array of inner product */
  Val    *midlist;		/* array of medians */
  Val    *temp;
  SCHED  *sched;		/* schedule for remapping */
  TTABLE *table;		/* translation table */
  TTABLE *init_ttable_with_proc();
  int	 comparedouble();

  /* initialize parameters */
  totalpart = (1 << level);

  /* allocate memory for necessary arrays*/
  index   = (int *) malloc ((ndata+PARTI_numnodes) * sizeof(int));
  proc    = (int *) malloc ((ndata+PARTI_numnodes) * sizeof(int));
  dot     = (double *) malloc ((ndata+PARTI_numnodes) * sizeof(double));
  weight  = (int *) malloc ((ndata+PARTI_numnodes) * sizeof(int));
  midlist = (Val *) malloc ((totalpart * PARTI_numnodes) * sizeof(Val));
  temp    = (Val *) malloc ((totalpart * PARTI_numnodes) * sizeof(Val));
  doublelist  = dot;

  /* put the coordinates into data array */
  for (i = 0; i <= ndata; i++) {
    index[i] = i;
    proc[i] = 0;
  }

  /* sort the arrays of each partition and find the median */
  for (i=0, ptr[0]=0; i < level; i++) {
    /* find the mass of partitions */
    npart = (1 << i);			/* number of partitions */
    ptr[npart] = ndata;    		/* number of nodes */
    for (j = 1, weight[0] = 0; j <= ndata; j++)
      weight[j] = weight[j-1] + load[index[j-1]];
    for (j = 0; j < npart; j++)
      gsize[j] = weight[ptr[j+1]] - weight[ptr[j]];
    addivec (level, gsize, npart, work);

    /* find the center of mass of each partition */
    for (j = 0; j < npart; j++)
      for (k = 0; k < ndim; k++)
        C[j][k] = 0.0;
    for (j = 0; j < npart; j++) {
      for (k = ptr[j]; k < ptr[j+1]; k++) {
	n = index[k];
	for (m = 0; m < ndim; m++)
          C[j][m] += load[n] * data[m][n];
      }
    }
    adddvec (level, C, npart*NDIM, dwork);
    for (j = 0; j < npart; j++)
      for (k = 0; k < ndim; k++)
    	C[j][k] /= gsize[j];
#ifdef	DEBUG
    if (PARTI_mynode == 0)
      printf ("Center of mass: %f %f %f\n", C[0][0], C[0][1], C[0][2]);
#endif

    /* calculate the moment of inertial tensor */
    for (j = 0; j < npart; j++)
      for (k = 0; k < NTENSOR; k++)
    	M[j][k] = 0.0;
    for (j = 0; j < npart; j++) {
      for (k = ptr[j]; k < ptr[j+1]; k++) {
	n = index[k];
	for (no = 0, l = 0; l < ndim; l++)
	  for (m = l; m < ndim; no++, m++)
            M[j][no] += load[n] * (data[l][n]-C[j][l]) * (data[m][n]-C[j][m]);
      }
    }
    adddvec (level, M, npart*NTENSOR, dwork);
#ifdef	DEBUG
    if (PARTI_mynode == 0) {
      printf ("Moment of Inertial Tensor:\n");
      printf ("\t%f\t%f\t%f\n", M[0][0], M[0][1], M[0][2]);
      printf ("\t%f\t%f\t%f\n", M[0][1], M[0][3], M[0][4]);
      printf ("\t%f\t%f\t%f\n", M[0][2], M[0][4], M[0][5]);
    }
#endif

    /* find the eigenvalues of eigenvectors */
    for (j = 0; j < npart; j++)
      find_eigenvector (ndim, &M[j][0], &V[j][0]);
#ifdef	DEBUG
    if (PARTI_mynode == 0)
      printf ("Eigenvector: %f %f %f\n", V[0][0], V[0][1], V[0][2]);
#endif

    /* projection (inner product) of each element to eigenvector */
    for (j = 0; j < npart; j++) {
      for (k = ptr[j]; k < ptr[j+1]; k++) {
        n = index[k];
	for (dot[n] = 0, m = 0; m < ndim; m++)
          dot[n] += V[j][m] * (data[m][n] - C[j][m]);
      }
    }

    /* sort the dot product array of each partition */
    for (j = 0; j < npart; j++)
      qsort (&index[ptr[j]], ptr[j+1]-ptr[j], sizeof(int), comparedouble);

    /* calculate the weight */
    for (j = 1, weight[0] = 0; j <= ndata; j++)
      weight[j] = weight[j-1] + load[index[j-1]];

    /* apply the method median of medians to find 
       the median of the distributed array */
    find_median_weight (level, npart, ptr, dot, weight, index, proc, midlist, temp);
  }

  /* put the result of partition to maparray */
  for (i = 0; i < ndata; i++)
    maparray[index[i]] = proc[index[i]];

  /* release memory */
  free (index);
  free (proc);
  free (dot);
  free (weight);
  free (temp);
  free (midlist);
}

/* when the number of processor is large, the calls to WeighInerBisecMap()
   will be changed to WeighInerBisec() to reduce execution time. */
void
WeighRecInerBisecRemap (level, maparray, load, ndata, ndim, data)
int	level;			/* level of binary disection */
int     *maparray;              /* result of partitioning */
int	*load;			/* loads of elements */
int     ndata;                  /* number of nodes */
int	ndim;			/* number of dimenstions */
double  *data[];		/* coordinates */

{
    int     start;          /* start index of local data */
    int     newndata;       /* number of data after partitioning */
    int     nlist[MAXPART]; /* list of ndata */
    int     temp[MAXPART];  /* temporary array */
    int     *myindex;       /* indexes of local elements */
    int     totalsize, localsize;
    int	    *newload;
    int     *proc, *local;
    int     i, j;
    double  **newdata;
    TTABLE  *table;
    SCHED   *sched;

    /* find the global size */
    for (i = 0; i < PARTI_numnodes; i++)
        nlist[i] = 0;
    nlist[PARTI_mynode] = ndata;
    PARTI_gisum (nlist, PARTI_numnodes, temp);
    for (localsize = nlist[0], i = 1; i < PARTI_numnodes; i++) {
	if (nlist[i] > localsize)
	    localsize = nlist[i];
        nlist[i] += nlist[i-1];
    }
    totalsize = nlist[PARTI_numnodes-1];
    localsize = ((localsize << 1) > totalsize)? totalsize: (localsize << 1);

    /* allocate memory */
    myindex = (int *) malloc (localsize * sizeof(int));
    newload = (int *) malloc (localsize * sizeof(int));
    newdata = (double **) malloc (sizeof(double *) * ndim);
    for (i = 0; i < ndim; i++)
        newdata[i] = (double *) malloc (localsize * sizeof(double));

    /* generate indexes */
    start = (PARTI_mynode) ? nlist[PARTI_mynode-1]: 0;
    for (i = 0; i < ndata; i++) {
        myindex[i] = start + i;
	newload[i] = load[i];
    }

    /* build translation table */
    table = build_translation_table (BLOCKED, myindex, ndata);

    /* copy data arrays to new arrays */
    for (i = 0; i < ndim; i++)
        for (j = 0; j < ndata; j++) {
    	    newdata[i][j] = data[i][j];
    }

    /* call the binary disection */
    newndata = ndata;
    WeighRecInerBisec (level, REMAPLEVEL, myindex, newload, &newndata, 3, newdata);

    /* free allocated memory */
    for (i = 0; i < ndim; i++)
        free (newdata[i]);
    free (newdata);

    /* generate a schedule */
    proc  = (int *) malloc (newndata * sizeof(int));
    local = (int *) malloc (newndata * sizeof(int));
    dereference (table, myindex, local, proc, newndata);
    sched = schedule (local, proc, newndata, 1);

    /* produce maparray */
    for (i = 0; i < newndata; i++)
        proc[i] = PARTI_mynode;
    PARTI_scatter (sched, proc, maparray, sizeof(int), NULL);

    /* free allocated memory */
    free (local);
    free (proc);
    free (newload);
    free (myindex);
    free_table (table);
    free_sched (sched);
}

/*
 * find the eigenvector of the largest eigenvalue of the inertial tensor
 */

find_eigenvector (ndim, tensor, vector)
int	ndim;				/* number of dimensions */
double	*tensor;			/* inertial tensor */
double	*vector;			/* eigenvector */

{
  double *M[NDIM];			/* matrix of inertial tensor */
  double E[NDIM];			/* eigenvalues of inertial tensor */
  double **V;				/* eigenvectors of eigenvalues */
  double temp;
  double **allocate_matrix();
  int	 i, n, t;

  /* set up the matrice */
  for (i = 0, n = 0, t = 0; i < ndim; i++) {
    t += i;
    M[i] = &tensor[n-t];
    n += ndim;
  }
  V = allocate_matrix(ndim, ndim, sizeof(double));

  /* find the eigenvalues and eigenvectors */
  jacobi (M, ndim, E, V, &n);

  /* find the larget eigenvalue */
  for (i = 1, temp = E[0], n = 0; i < ndim; i++)
    if (E[i] > temp) {
      temp = E[i];
      n = i;
    }

  /* eigenvector of the largest eigenvalue */
  for (i = 0; i < ndim; i++)
    vector[i] = V[i][n];

  /* free allocated memory */
  free_matrix (ndim, ndim, V);
}
